"""
Create an abstract class for the cryptographic functions
"""
from pypraware_crypto.interface import Crypto
from pypraware_crypto.helper import *
import configparser
import glob, os, hashlib
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes

class SHA(Crypto):
    def __init__(self, conf, name, metadata=None):
        print(name)
        # check the name and raise error otherwise
        if name not in ['sha256', 'sha384', 'sha512']:
            raise BaseException('Algorithm is not accepted, try in sha256, sha385, sha521')
        self.conf = conf
        self.hash_name = name
        self.btoken = bytes(conf['misp']['token'], encoding='ascii')
        # For matching (only token is kept from config file)
        if metadata is not None:
            metadata = metadata['crypto']
            self.hash_name = name

    def derive_key(self, bpassword, bsalt, attr_types):
        """
        Generate the key for encryption
        """
        name = self.hash_name
        if name == 'sha256':
            theHash = hashes.SHA256()
        elif name == 'sha384':
            theHash = hashes.SHA284()
        elif name == 'sha512':
            theHash = hashes.SHA512()

        digest = hashes.Hash(theHash, backend=default_backend())
        digest.update(bpassword + self.btoken + bsalt)
        return digest.finalize() 

    def create_rule(self, ioc, message):
        salt = os.urandom(hashlib.new(self.hash_name).digest_size)
        attr_types, password = get_types_values(ioc)
        dk = self.derive_key(password.encode('utf8'), salt, attr_types)
        return aes_create_rule(dk, message, attr_types, salt)

    def match(self, attributes, rule, queue):
        """
        Sometimes we don't need to decrypt the whole
        ciphertext to know if there is a match
        as it is the case here thanks to ctr mode
        """
        rule_attr = rule['attributes']
        match = False
        try:
            password = '||'.join([attributes[attr] for attr in rule_attr])
            attr_types = '||'.join(attr_type for attr_type in rule_attr)
            ciphertext = [rule['ciphertext-check'], rule['ciphertext']]
            dk = self.derive_key(password.encode('utf8'), rule['salt'], attr_types)
            match, plaintext = aes_match_rule(dk, password,\
                    rule['nonce'], ciphertext)
        except:
            pass # nothing to do

        if match:
            queue.put("IOC matched for: {}\nSecret Message (uuid-event id-date)\n===================================\n{}\n".format(attributes, plaintext.decode('utf-8')))



    def save_meta(self):
        meta = configparser.ConfigParser()
        meta['crypto'] = {}
        meta['crypto']['name'] = self.hash_name
        with open(self.conf['rules']['location'] + '/metadata', 'w') as config:
            meta.write(config)




"""
Create an abstract class for the cryptographic functions
Configuration must me in the configuration file
"""

class Crypto():
    def __init__(self, conf, metadata=None, cryptoName=''):
        """
        Metadata is used for matching
        """
        pass

    def create_rule(self, ioc, message):
        """
        Use the generated key and salt to 
        encrypt the message
        """
        pass

    def match(self, attributes, rule, queue):
        """
        attributes is the values to check
        rule is the rule against which it has to be check.
            for an example it can be a bloom filter
        queue is to send the output (easier to redirect only
            one output is further needed)
        """
        pass

    def save_meta(self):
        """
        Save metadata for the specific crypto system
        """
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Do not duplicate code in both readMisp and matchRules
The goal is first to normalize URLs and IP addresses
URLs:
    - standards normalization as done with the library
    - additionnal 'normalization' steps explained in the report
        (could transform the URL but better for matching)
IP: 
    - Standardize the IPv4, IPv6 and ranges
    - If the library throws an error, the value is added without modifications
        and a message appear in the console
"""
import re, urllib
from url_normalize import url_normalize
from ipaddress import ip_address, ip_network


def normalize(ioc):
    for attr_type in ioc:
        # url|uri|link are often misused
        # Thus they are considered to be the same
        if attr_type == 'url' or\
            attr_type == 'uri' or\
            attr_type == 'link':
                # just solve one specific case:
                if not '..org' in ioc[attr_type]:
                    try:
                        ioc[attr_type] = urlNorm(ioc[attr_type])
                    except Exception as e:
                        print('\nUrl normalisation error with: ' +\
                         ioc[attr_type] + '\n' + str(e) + '\n')
                        ioc[attr_type] = ioc[attr_type]
        elif attr_type == 'hostname':
                ioc[attr_type] = ioc[attr_type].lower()
        elif 'ip-' in attr_type:
            ioc[attr_type] = ipNorm(ioc[attr_type])
    return ioc


def ipNorm(ip):
    # This normalize IP
    try:
        if '/' in ip:
            return str(ip_network(ip))
        else:
            return str(ip_address(ip))
    except Exception as e:
        print('\nIP normalisation raised an error with this ip ' +\
         ip +'\n' + str(e) + '\n')
        return ip

directory_indexes = ['default.asp', 'index.html', 'index.php', 'index.shtml'\
                    'index.jsp', '\?']
def urlNorm(url):
    url = url_normalize(url)
    # removes fragment
    url = urllib.parse.urldefrag(url)[0]

    # remove index directories if it is at the end
    for index in directory_indexes:
        url = re.sub(index+ '$', '', url)
    
    # remove http https and www.
    url = re.sub("^https?://(www.)?", '', url)
    return url



def Crypto(name, conf, metadata=None):
    name = name.lower()
    if (name == 'pbkdf2'):
        from pypraware_crypto.pbkdf2 import Pbkdf2
        return Pbkdf2(conf, metadata)
    elif (name == 'bcrypt'):
        from pypraware_crypto.bcrypt import Bcrypt
        return Bcrypt(conf, metadata)
    elif (name == 'hkdf'):
        from pypraware_crypto.hkdf import HKDF
        return HKDF(conf, metadata)
    elif (name == 'bloom_filter'):
        from pypraware_crypto.bloom_filter import Bloom_filter
        return Bloom_filter(conf, metadata)
    elif (name.startswith('bloomy_')):
        from pypraware_crypto.bloomy import Bloomy
        return Bloomy(conf, metadata, name[7:])
    elif (name.startswith('sha')):
        from pypraware_crypto.sha import SHA
        return SHA(conf, name, metadata)
    else:
        print('Not recognized')



"""
Create an abstract class for the cryptographic functions
"""
from pypraware_crypto.interface import Crypto
from pypraware_crypto.helper import *
import configparser
import glob, hashlib, os

class Pbkdf2(Crypto):
    def __init__(self, conf, metadata=None):
        self.conf = conf
        self.hash_name = conf['pbkdf2']['hash_name']
        self.dklen = int(conf['pbkdf2']['dklen'])
        self.btoken = bytes(conf['misp']['token'], encoding='ascii')
        self.iterations = int(self.conf['pbkdf2']['iterations'])
        self.ipiterations = int(self.conf['pbkdf2']['ipiterations'])
        # For matching (only token is kept from config file)
        if metadata is not None:
            metadata = metadata['crypto']
            self.hash_name = metadata['hash_name']
            self.dklen = int(metadata['dklen'])
            self.iterations = int(metadata['iterations'])
            self.ipiterations = int(metadata['ipiterations'])

    def derive_key(self, bpassword, bsalt, attr_types):
        """
        Generate the key for encryption
        """
        it = 1
        if 'ip' in attr_types:
            it = self.ipiterations
        else:
            it = self.iterations
        return hashlib.pbkdf2_hmac(self.hash_name, bpassword + self.btoken, bsalt, it, dklen=self.dklen)

    def create_rule(self, ioc, message):
        salt = os.urandom(hashlib.new(self.hash_name).digest_size)
        attr_types, password = get_types_values(ioc)
        dk = self.derive_key(password.encode('utf8'), salt, attr_types)
        return aes_create_rule(dk, message, attr_types, salt)

    def match(self, attributes, rule, queue):
        """
        Sometimes we don't need to decrypt the whole
        ciphertext to know if there is a match
        as it is the case here thanks to ctr mode
        """
        rule_attr = rule['attributes']
        match = False
        try:
            password = '||'.join([attributes[attr] for attr in rule_attr])
            attr_types = '||'.join(attr_type for attr_type in rule_attr)
            ciphertext = [rule['ciphertext-check'], rule['ciphertext']]
            dk = self.derive_key(password.encode('utf8'), rule['salt'], attr_types)
            match, plaintext = aes_match_rule(dk, password,\
                    rule['nonce'], ciphertext)
        except:
            pass # nothing to do

        if match:
            queue.put("IOC matched for: {}\nSecret Message (uuid-event id-date)\n===================================\n{}\n".format(attributes, plaintext.decode('utf-8')))



    def save_meta(self):
        meta = configparser.ConfigParser()
        meta['crypto'] = {}
        meta['crypto']['name'] = 'pbkdf2' 
        meta['crypto']['hash_name'] = self.conf['pbkdf2']['hash_name']
        meta['crypto']['dklen'] = self.conf['pbkdf2']['dklen'] # AES block size
        meta['crypto']['iterations'] = str(self.iterations)
        meta['crypto']['ipiterations'] = str(self.ipiterations)
        with open(self.conf['rules']['location'] + '/metadata', 'w') as config:
            meta.write(config)




"""
The goal of this implementation is to mix a bloom filter
with false positive rate defined in the configuration file in order to :
    -> make matching system faster
    -> Do not give too many information only with the bloomFilter
with a standard choosen key derivation function
"""
from pypraware_crypto.interface import Crypto
from pypraware_crypto.bloom_filter import Bloom_filter as BF
from pypraware_crypto.choose_crypto import Crypto as ChooseCrypto
from configparser import ConfigParser

class Bloomy(Crypto):
    def __init__(self, conf, metadata=None, cryptoName=''):
        self.conf = conf
        self.Crypto = ChooseCrypto(cryptoName, conf, metadata)
        # Set up bloom
        self.bloom = BF(conf, metadata=metadata, rate=conf['bloomy']['fp_rate'])
        # Make it faster without modifying 
        self.cacheVal = ""
        self.cacheBool = True

    def create_rule(self, ioc, message):
        self.bloom.create_rule(ioc, message)
        return self.Crypto.create_rule(ioc, message)

    def match(self, attributes, rule, queue):
        if self.cacheVal != attributes:
            self.cacheVal = attributes
            if len(self.bloom.check(attributes, rule))>0:
                self.cacheBool = True
                self.Crypto.match(attributes, rule, queue)
            else:
                self.cacheBool = False
        else:
            if self.cacheBool == True:
                self.Crypto.match(attributes, rule, queue)

    def save_meta(self):
        conf = self.conf
        # Create bloom filter
        self.bloom.write_bloom()
        # Save metadata
        self.Crypto.save_meta()
        # Add bloomy_
        metaParser = ConfigParser()
        metaParser.read(conf['rules']['location'] + '/metadata')
        metadata = metaParser._sections
        cryptoName = 'bloomy_' + metadata['crypto']['name']
        metaParser['crypto']['name'] = cryptoName
        with open(conf['rules']['location'] + '/metadata', 'w') as config:
            metaParser.write(config)




"""A setuptools based setup module.(Pypa)
"""

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pypraware',

    version='0.1.9',

    description='Privacy aware library for MISP module',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/MISP/pypraware',

    # Author details
    author='Charles Jacquet',
    author_email='charles.jacquet7@gmail.com',

    # Choose your license
    license='GNU Affero General Public License v3',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU Affero General Public License v3',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    # What does your project relate to?
    keywords='MISP module privacy',

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    packages=["pypraware_crypto", "pypraware_normalize"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['configparser', 'cryptography', 'bcrypt', 'bitarray', 'url_normalize'],

)

